import time

from flask import json

from authlib.oauth2.rfc6749.grants import RefreshTokenGrant as _RefreshTokenGrant

from .models import Client
from .models import Token
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class RefreshTokenGrant(_RefreshTokenGrant):
    def authenticate_refresh_token(self, refresh_token):
        item = Token.query.filter_by(refresh_token=refresh_token).first()
        if item and item.is_refresh_token_active():
            return item

    def authenticate_user(self, credential):
        return db.session.get(User, credential.user_id)

    def revoke_old_credential(self, credential):
        now = int(time.time())
        credential.access_token_revoked_at = now
        credential.refresh_token_revoked_at = now
        db.session.add(credential)
        db.session.commit()


class RefreshTokenTest(TestCase):
    def prepare_data(self, grant_type="refresh_token"):
        server = create_authorization_server(self.app)
        server.register_grant(RefreshTokenGrant)

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="refresh-client",
            client_secret="refresh-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "grant_types": [grant_type],
                "redirect_uris": ["http://localhost/authorized"],
            }
        )
        db.session.add(client)
        db.session.commit()

    def create_token(self, scope="profile", user_id=1):
        token = Token(
            user_id=user_id,
            client_id="refresh-client",
            token_type="bearer",
            access_token="a1",
            refresh_token="r1",
            scope=scope,
            expires_in=3600,
        )
        db.session.add(token)
        db.session.commit()

    def test_invalid_client(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "foo",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("invalid-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "foo",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("refresh-client", "invalid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "foo",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_invalid_refresh_token(self):
        self.prepare_data()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")
        self.assertIn("Missing", resp["error_description"])

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "foo",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

    def test_invalid_scope(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_scope")

    def test_invalid_scope_none(self):
        self.prepare_data()
        self.create_token(scope=None)
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_scope")

    def test_invalid_user(self):
        self.prepare_data()
        self.create_token(user_id=5)
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

    def test_invalid_grant_type(self):
        self.prepare_data(grant_type="invalid")
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_authorize_token_no_scope(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_authorize_token_scope(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_revoke_old_credential(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 400)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})

        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("refresh-client", "refresh-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("r-refresh_token.1.", resp["access_token"])



from flask import current_app
from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import url_encode
from authlib.common.urls import urlparse
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core.grants import OpenIDCode as _OpenIDCode
from tests.util import read_file_path

from .models import Client
from .models import CodeGrantMixin
from .models import User
from .models import db
from .models import exists_nonce
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class OpenIDCode(_OpenIDCode):
    def get_jwt_config(self, grant):
        key = current_app.config["OAUTH2_JWT_KEY"]
        alg = current_app.config["OAUTH2_JWT_ALG"]
        iss = current_app.config["OAUTH2_JWT_ISS"]
        return dict(key=key, alg=alg, iss=iss, exp=3600)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)


class BaseTestCase(TestCase):
    def config_app(self):
        self.app.config.update(
            {
                "OAUTH2_JWT_ISS": "Authlib",
                "OAUTH2_JWT_KEY": "secret",
                "OAUTH2_JWT_ALG": "HS256",
            }
        )

    def prepare_data(self):
        self.config_app()
        server = create_authorization_server(self.app)
        server.register_grant(AuthorizationCodeGrant, [OpenIDCode()])

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

        client = Client(
            user_id=user.id,
            client_id="code-client",
            client_secret="code-secret",
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b"],
                "scope": "openid profile address",
                "response_types": ["code"],
                "grant_types": ["authorization_code"],
            }
        )
        db.session.add(client)
        db.session.commit()


class OpenIDCodeTest(BaseTestCase):
    def test_authorize_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "code",
                "client_id": "code-client",
                "state": "bar",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)

        claims = jwt.decode(
            resp["id_token"],
            "secret",
            claims_cls=CodeIDToken,
            claims_options={"iss": {"value": "Authlib"}},
        )
        claims.validate()

    def test_pure_code_flow(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "code",
                "client_id": "code-client",
                "state": "bar",
                "scope": "profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertNotIn("id_token", resp)

    def test_nonce_replay(self):
        self.prepare_data()
        data = {
            "response_type": "code",
            "client_id": "code-client",
            "user_id": "1",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://a.b",
        }
        rv = self.client.post("/oauth/authorize", data=data)
        self.assertIn("code=", rv.location)

        rv = self.client.post("/oauth/authorize", data=data)
        self.assertIn("error=", rv.location)

    def test_prompt(self):
        self.prepare_data()
        params = [
            ("response_type", "code"),
            ("client_id", "code-client"),
            ("state", "bar"),
            ("nonce", "abc"),
            ("scope", "openid profile"),
            ("redirect_uri", "https://a.b"),
        ]
        query = url_encode(params)
        rv = self.client.get("/oauth/authorize?" + query)
        self.assertEqual(rv.data, b"login")

        query = url_encode(params + [("user_id", "1")])
        rv = self.client.get("/oauth/authorize?" + query)
        self.assertEqual(rv.data, b"ok")

        query = url_encode(params + [("prompt", "login")])
        rv = self.client.get("/oauth/authorize?" + query)
        self.assertEqual(rv.data, b"login")

        query = url_encode(params + [("user_id", "1"), ("prompt", "login")])
        rv = self.client.get("/oauth/authorize?" + query)
        self.assertEqual(rv.data, b"login")


class RSAOpenIDCodeTest(BaseTestCase):
    def config_app(self):
        self.app.config.update(
            {
                "OAUTH2_JWT_ISS": "Authlib",
                "OAUTH2_JWT_KEY": read_file_path("jwk_private.json"),
                "OAUTH2_JWT_ALG": "RS256",
            }
        )

    def get_validate_key(self):
        return read_file_path("jwk_public.json")

    def test_authorize_token(self):
        # generate refresh token
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "code",
                "client_id": "code-client",
                "state": "bar",
                "scope": "openid profile",
                "redirect_uri": "https://a.b",
                "user_id": "1",
            },
        )
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.assertEqual(params["state"], "bar")

        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://a.b",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)

        claims = jwt.decode(
            resp["id_token"],
            self.get_validate_key(),
            claims_cls=CodeIDToken,
            claims_options={"iss": {"value": "Authlib"}},
        )
        claims.validate()


class JWKSOpenIDCodeTest(RSAOpenIDCodeTest):
    def config_app(self):
        self.app.config.update(
            {
                "OAUTH2_JWT_ISS": "Authlib",
                "OAUTH2_JWT_KEY": read_file_path("jwks_private.json"),
                "OAUTH2_JWT_ALG": "PS256",
            }
        )

    def get_validate_key(self):
        return read_file_path("jwks_public.json")


class ECOpenIDCodeTest(RSAOpenIDCodeTest):
    def config_app(self):
        self.app.config.update(
            {
                "OAUTH2_JWT_ISS": "Authlib",
                "OAUTH2_JWT_KEY": read_file_path("secp521r1-private.json"),
                "OAUTH2_JWT_ALG": "ES512",
            }
        )

    def get_validate_key(self):
        return read_file_path("secp521r1-public.json")


class PEMOpenIDCodeTest(RSAOpenIDCodeTest):
    def config_app(self):
        self.app.config.update(
            {
                "OAUTH2_JWT_ISS": "Authlib",
                "OAUTH2_JWT_KEY": read_file_path("rsa_private.pem"),
                "OAUTH2_JWT_ALG": "RS256",
            }
        )

    def get_validate_key(self):
        return read_file_path("rsa_public.pem")



from flask_sqlalchemy import SQLAlchemy

from authlib.integrations.sqla_oauth2 import OAuth2AuthorizationCodeMixin
from authlib.integrations.sqla_oauth2 import OAuth2ClientMixin
from authlib.integrations.sqla_oauth2 import OAuth2TokenMixin
from authlib.oidc.core import UserInfo

db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True, nullable=False)

    def get_user_id(self):
        return self.id

    def check_password(self, password):
        return password != "wrong"

    def generate_user_info(self, scopes):
        profile = {"sub": str(self.id), "name": self.username}
        return UserInfo(profile)


class Client(db.Model, OAuth2ClientMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")


class AuthorizationCode(db.Model, OAuth2AuthorizationCodeMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, nullable=False)

    @property
    def user(self):
        return db.session.get(User, self.user_id)


class Token(db.Model, OAuth2TokenMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")

    def is_refresh_token_active(self):
        return not self.refresh_token_revoked_at


class CodeGrantMixin:
    def query_authorization_code(self, code, client):
        item = AuthorizationCode.query.filter_by(
            code=code, client_id=client.client_id
        ).first()
        if item and not item.is_expired():
            return item

    def delete_authorization_code(self, authorization_code):
        db.session.delete(authorization_code)
        db.session.commit()

    def authenticate_user(self, authorization_code):
        return db.session.get(User, authorization_code.user_id)


def save_authorization_code(code, request):
    client = request.client
    auth_code = AuthorizationCode(
        code=code,
        client_id=client.client_id,
        redirect_uri=request.redirect_uri,
        scope=request.scope,
        nonce=request.data.get("nonce"),
        user_id=request.user.id,
        code_challenge=request.data.get("code_challenge"),
        code_challenge_method=request.data.get("code_challenge_method"),
    )
    db.session.add(auth_code)
    db.session.commit()
    return auth_code


def exists_nonce(nonce, req):
    exists = AuthorizationCode.query.filter_by(
        client_id=req.client_id, nonce=nonce
    ).first()
    return bool(exists)



from flask import json

from authlib.common.urls import add_params_to_uri
from authlib.oauth2.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _PasswordGrant,
)
from authlib.oidc.core import OpenIDToken

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class IDToken(OpenIDToken):
    def get_jwt_config(self, grant):
        return {
            "iss": "Authlib",
            "key": "secret",
            "alg": "HS256",
        }

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)


class PasswordGrant(_PasswordGrant):
    def authenticate_user(self, username, password):
        user = User.query.filter_by(username=username).first()
        if user.check_password(password):
            return user


class PasswordTest(TestCase):
    def prepare_data(self, grant_type="password", extensions=None):
        server = create_authorization_server(self.app)
        server.register_grant(PasswordGrant, extensions)
        self.server = server

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="password-client",
            client_secret="password-secret",
        )
        client.set_client_metadata(
            {
                "scope": "openid profile",
                "grant_types": [grant_type],
                "redirect_uris": ["http://localhost/authorized"],
            }
        )
        db.session.add(client)
        db.session.commit()

    def test_invalid_client(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("password-client", "invalid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_invalid_scope(self):
        self.prepare_data()
        self.server.scopes_supported = ["profile"]
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
                "scope": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_scope")

    def test_invalid_request(self):
        self.prepare_data()
        headers = self.create_basic_header("password-client", "password-secret")

        rv = self.client.get(
            add_params_to_uri(
                "/oauth/token",
                {
                    "grant_type": "password",
                },
            ),
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_grant_type")

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "wrong",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

    def test_invalid_grant_type(self):
        self.prepare_data(grant_type="invalid")
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_authorize_token(self):
        self.prepare_data()
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
        self.prepare_data()
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("p-password.1.", resp["access_token"])

    def test_custom_expires_in(self):
        self.app.config.update({"OAUTH2_TOKEN_EXPIRES_IN": {"password": 1800}})
        self.prepare_data()
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertEqual(resp["expires_in"], 1800)

    def test_id_token_extension(self):
        self.prepare_data(extensions=[IDToken()])
        headers = self.create_basic_header("password-client", "password-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "password",
                "username": "foo",
                "password": "ok",
                "scope": "openid profile",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("id_token", resp)



from typing import Optional

from authlib.common.urls import add_params_to_uri


class IssuerParameter:
    def __call__(self, grant):
        grant.register_hook(
            "after_authorization_response",
            self.add_issuer_parameter,
        )

    def add_issuer_parameter(self, hook_type: str, response):
        if self.get_issuer():
            # RFC9207 §2
            # In authorization responses to the client, including error responses,
            # an authorization server supporting this specification MUST indicate
            # its identity by including the iss parameter in the response.

            new_location = add_params_to_uri(
                response.location, {"iss": self.get_issuer()}
            )
            response.location = new_location

    def get_issuer(self) -> Optional[str]:
        """Return the issuer URL.
        Developers MAY implement this method if they want to support :rfc:`RFC9207 <9207>`::

            def get_issuer(self) -> str:
                return "https://auth.example.org"
        """
        return None



from flask import json

from authlib.oauth2.rfc6749.grants import ClientCredentialsGrant

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class ClientCredentialsTest(TestCase):
    def prepare_data(self, grant_type="client_credentials"):
        server = create_authorization_server(self.app)
        server.register_grant(ClientCredentialsGrant)
        self.server = server

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="credential-client",
            client_secret="credential-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
                "grant_types": [grant_type],
            }
        )
        db.session.add(client)
        db.session.commit()

    def test_invalid_client(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("credential-client", "invalid-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_invalid_grant_type(self):
        self.prepare_data(grant_type="invalid")
        headers = self.create_basic_header("credential-client", "credential-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_invalid_scope(self):
        self.prepare_data()
        self.server.scopes_supported = ["profile"]
        headers = self.create_basic_header("credential-client", "credential-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "scope": "invalid",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_scope")

    def test_authorize_token(self):
        self.prepare_data()
        headers = self.create_basic_header("credential-client", "credential-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})

        self.prepare_data()
        headers = self.create_basic_header("credential-client", "credential-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)
        self.assertIn("c-client_credentials.", resp["access_token"])



from authlib.common.errors import AuthlibHTTPError
from authlib.common.urls import add_params_to_uri


class OAuth2Error(AuthlibHTTPError):
    def __init__(
        self,
        description=None,
        uri=None,
        status_code=None,
        state=None,
        redirect_uri=None,
        redirect_fragment=False,
        error=None,
    ):
        super().__init__(error, description, uri, status_code)
        self.state = state
        self.redirect_uri = redirect_uri
        self.redirect_fragment = redirect_fragment

    def get_body(self):
        """Get a list of body."""
        error = super().get_body()
        if self.state:
            error.append(("state", self.state))
        return error

    def __call__(self, uri=None):
        if self.redirect_uri:
            params = self.get_body()
            loc = add_params_to_uri(self.redirect_uri, params, self.redirect_fragment)
            return 302, "", [("Location", loc)]
        return super().__call__(uri=uri)



from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc9207 import IssuerParameter as _IssuerParameter

from .models import Client
from .models import CodeGrantMixin
from .models import User
from .models import db
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class IssuerParameter(_IssuerParameter):
    def get_issuer(self) -> str:
        return "https://auth.test"


class RFC9207AuthorizationCodeTest(TestCase):
    LAZY_INIT = False

    def prepare_data(
        self,
        is_confidential=True,
        response_type="code",
        grant_type="authorization_code",
        token_endpoint_auth_method="client_secret_basic",
        rfc9207=True,
    ):
        server = create_authorization_server(self.app, self.LAZY_INIT)
        extensions = [IssuerParameter()] if rfc9207 else []
        server.register_grant(AuthorizationCodeGrant, extensions=extensions)
        self.server = server

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

        if is_confidential:
            client_secret = "code-secret"
        else:
            client_secret = ""
        client = Client(
            user_id=user.id,
            client_id="code-client",
            client_secret=client_secret,
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b"],
                "scope": "profile address",
                "token_endpoint_auth_method": token_endpoint_auth_method,
                "response_types": [response_type],
                "grant_types": grant_type.splitlines(),
            }
        )
        self.authorize_url = "/oauth/authorize?response_type=code&client_id=code-client"
        db.session.add(client)
        db.session.commit()

    def test_rfc9207_enabled_success(self):
        """Check that when RFC9207 is implemented,
        the authorization response has an ``iss`` parameter."""

        self.prepare_data(rfc9207=True)
        url = self.authorize_url + "&state=bar"
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("iss=https%3A%2F%2Fauth.test", rv.location)

    def test_rfc9207_disabled_success_no_iss(self):
        """Check that when RFC9207 is not implemented,
        the authorization response contains no ``iss`` parameter."""

        self.prepare_data(rfc9207=False)
        url = self.authorize_url + "&state=bar"
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertNotIn("iss=", rv.location)

    def test_rfc9207_enabled_error(self):
        """Check that when RFC9207 is implemented,
        the authorization response has an ``iss`` parameter,
        even when an error is returned."""

        self.prepare_data(rfc9207=True)
        rv = self.client.post(self.authorize_url)
        self.assertIn("error=access_denied", rv.location)
        self.assertIn("iss=https%3A%2F%2Fauth.test", rv.location)

    def test_rfc9207_disbled_error_no_iss(self):
        """Check that when RFC9207 is not implemented,
        the authorization response contains no ``iss`` parameter,
        even when an error is returned."""

        self.prepare_data(rfc9207=False)
        rv = self.client.post(self.authorize_url)
        self.assertIn("error=access_denied", rv.location)
        self.assertNotIn("iss=", rv.location)



import json

from authlib.integrations.django_oauth2 import RevocationEndpoint

from .models import Client
from .models import OAuth2Token
from .models import User
from .oauth2_server import TestCase

ENDPOINT_NAME = RevocationEndpoint.ENDPOINT_NAME


class RevocationEndpointTest(TestCase):
    def create_server(self):
        server = super().create_server()
        server.register_endpoint(RevocationEndpoint)
        return server

    def prepare_client(self):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            token_endpoint_auth_method="client_secret_basic",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def prepare_token(self, scope="profile", user_id=1):
        token = OAuth2Token(
            user_id=user_id,
            client_id="client",
            token_type="bearer",
            access_token="a1",
            refresh_token="r1",
            scope=scope,
            expires_in=3600,
        )
        token.save()

    def test_invalid_client(self):
        server = self.create_server()
        request = self.factory.post("/oauth/revoke")
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        request = self.factory.post("/oauth/revoke", HTTP_AUTHORIZATION="invalid token")
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        request = self.factory.post(
            "/oauth/revoke",
            HTTP_AUTHORIZATION=self.create_basic_auth("invalid", "secret"),
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        request = self.factory.post(
            "/oauth/revoke",
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "invalid"),
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

    def test_invalid_token(self):
        server = self.create_server()
        self.prepare_client()
        self.prepare_token()
        auth_header = self.create_basic_auth("client", "secret")

        request = self.factory.post("/oauth/revoke", HTTP_AUTHORIZATION=auth_header)
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")

        # case 1
        request = self.factory.post(
            "/oauth/revoke",
            data={"token": "invalid-token"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        self.assertEqual(resp.status_code, 200)

        # case 2
        request = self.factory.post(
            "/oauth/revoke",
            data={
                "token": "a1",
                "token_type_hint": "unsupported_token_type",
            },
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "unsupported_token_type")

        # case 3
        request = self.factory.post(
            "/oauth/revoke",
            data={
                "token": "a1",
                "token_type_hint": "refresh_token",
            },
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        self.assertEqual(resp.status_code, 200)

    def test_revoke_token_with_hint(self):
        self.prepare_client()
        self.prepare_token()
        self.revoke_token({"token": "a1", "token_type_hint": "access_token"})
        self.revoke_token({"token": "r1", "token_type_hint": "refresh_token"})

    def test_revoke_token_without_hint(self):
        self.prepare_client()
        self.prepare_token()
        self.revoke_token({"token": "a1"})
        self.revoke_token({"token": "r1"})

    def revoke_token(self, data):
        server = self.create_server()
        auth_header = self.create_basic_auth("client", "secret")

        request = self.factory.post(
            "/oauth/revoke",
            data=data,
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_endpoint_response(ENDPOINT_NAME, request)
        self.assertEqual(resp.status_code, 200)



from authlib.common.urls import add_params_to_uri
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.jose import JsonWebToken
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core.grants import OpenIDImplicitGrant as _OpenIDImplicitGrant

from .models import Client
from .models import User
from .models import db
from .models import exists_nonce
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class OpenIDImplicitGrant(_OpenIDImplicitGrant):
    def get_jwt_config(self):
        return dict(key="secret", alg="HS256", iss="Authlib", exp=3600)

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)


class ImplicitTest(TestCase):
    def prepare_data(self):
        server = create_authorization_server(self.app)
        server.register_grant(OpenIDImplicitGrant)

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="implicit-client",
            client_secret="",
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b/c"],
                "scope": "openid profile",
                "token_endpoint_auth_method": "none",
                "response_types": ["id_token", "id_token token"],
            }
        )
        self.authorize_url = (
            "/oauth/authorize?response_type=token&client_id=implicit-client"
        )
        db.session.add(client)
        db.session.commit()

    def validate_claims(self, id_token, params):
        jwt = JsonWebToken(["HS256"])
        claims = jwt.decode(
            id_token, "secret", claims_cls=ImplicitIDToken, claims_params=params
        )
        claims.validate()

    def test_consent_view(self):
        self.prepare_data()
        rv = self.client.get(
            add_params_to_uri(
                "/oauth/authorize",
                {
                    "response_type": "id_token",
                    "client_id": "implicit-client",
                    "scope": "openid profile",
                    "state": "foo",
                    "redirect_uri": "https://a.b/c",
                    "user_id": "1",
                },
            )
        )
        self.assertIn(b"error=invalid_request", rv.data)
        self.assertIn(b"nonce", rv.data)

    def test_require_nonce(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn("error=invalid_request", rv.location)
        self.assertIn("nonce", rv.location)

    def test_missing_openid_in_scope(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token token",
                "client_id": "implicit-client",
                "scope": "profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn("error=invalid_scope", rv.location)

    def test_denied(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
            },
        )
        self.assertIn("error=access_denied", rv.location)

    def test_authorize_access_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token token",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn("access_token=", rv.location)
        self.assertIn("id_token=", rv.location)
        self.assertIn("state=bar", rv.location)
        params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
        self.validate_claims(params["id_token"], params)

    def test_authorize_id_token(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn("id_token=", rv.location)
        self.assertIn("state=bar", rv.location)
        params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
        self.validate_claims(params["id_token"], params)

    def test_response_mode_query(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token",
                "response_mode": "query",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn("id_token=", rv.location)
        self.assertIn("state=bar", rv.location)
        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        self.validate_claims(params["id_token"], params)

    def test_response_mode_form_post(self):
        self.prepare_data()
        rv = self.client.post(
            "/oauth/authorize",
            data={
                "response_type": "id_token",
                "response_mode": "form_post",
                "client_id": "implicit-client",
                "scope": "openid profile",
                "state": "bar",
                "nonce": "abc",
                "redirect_uri": "https://a.b/c",
                "user_id": "1",
            },
        )
        self.assertIn(b'name="id_token"', rv.data)
        self.assertIn(b'name="state"', rv.data)



import unittest

from authlib.jose import JsonWebEncryption
from authlib.jose import OctKey
from authlib.jose.drafts import register_jwe_draft

register_jwe_draft(JsonWebEncryption)


class ChaCha20Test(unittest.TestCase):
    def test_dir_alg_c20p(self):
        jwe = JsonWebEncryption()
        key = OctKey.generate_key(256, is_private=True)
        protected = {"alg": "dir", "enc": "C20P"}
        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

        key2 = OctKey.generate_key(128, is_private=True)
        self.assertRaises(ValueError, jwe.deserialize_compact, data, key2)

        self.assertRaises(ValueError, jwe.serialize_compact, protected, b"hello", key2)

    def test_dir_alg_xc20p(self):
        jwe = JsonWebEncryption()
        key = OctKey.generate_key(256, is_private=True)
        protected = {"alg": "dir", "enc": "XC20P"}
        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

        key2 = OctKey.generate_key(128, is_private=True)
        self.assertRaises(ValueError, jwe.deserialize_compact, data, key2)

        self.assertRaises(ValueError, jwe.serialize_compact, protected, b"hello", key2)

    def test_xc20p_content_encryption_decryption(self):
        # https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-xchacha-03#appendix-A.3.1
        enc = JsonWebEncryption.ENC_REGISTRY["XC20P"]

        plaintext = bytes.fromhex(
            "4c616469657320616e642047656e746c656d656e206f662074686520636c6173"
            + "73206f66202739393a204966204920636f756c64206f6666657220796f75206f"
            + "6e6c79206f6e652074697020666f7220746865206675747572652c2073756e73"
            + "637265656e20776f756c642062652069742e"
        )
        aad = bytes.fromhex("50515253c0c1c2c3c4c5c6c7")
        key = bytes.fromhex(
            "808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9f"
        )
        iv = bytes.fromhex("404142434445464748494a4b4c4d4e4f5051525354555657")

        ciphertext, tag = enc.encrypt(plaintext, aad, iv, key)
        self.assertEqual(
            ciphertext,
            bytes.fromhex(
                "bd6d179d3e83d43b9576579493c0e939572a1700252bfaccbed2902c21396cbb"
                + "731c7f1b0b4aa6440bf3a82f4eda7e39ae64c6708c54c216cb96b72e1213b452"
                + "2f8c9ba40db5d945b11b69b982c1bb9e3f3fac2bc369488f76b2383565d3fff9"
                + "21f9664c97637da9768812f615c68b13b52e"
            ),
        )
        self.assertEqual(tag, bytes.fromhex("c0875924c1c7987947deafd8780acf49"))

        decrypted_plaintext = enc.decrypt(ciphertext, aad, iv, tag, key)
        self.assertEqual(decrypted_plaintext, plaintext)



import time

from flask import json

from authlib.oauth2.rfc8628 import (
    DeviceAuthorizationEndpoint as _DeviceAuthorizationEndpoint,
)
from authlib.oauth2.rfc8628 import DeviceCodeGrant as _DeviceCodeGrant
from authlib.oauth2.rfc8628 import DeviceCredentialDict

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server

device_credentials = {
    "valid-device": {
        "client_id": "client",
        "expires_in": 1800,
        "user_code": "code",
    },
    "expired-token": {
        "client_id": "client",
        "expires_in": -100,
        "user_code": "none",
    },
    "invalid-client": {
        "client_id": "invalid",
        "expires_in": 1800,
        "user_code": "none",
    },
    "denied-code": {
        "client_id": "client",
        "expires_in": 1800,
        "user_code": "denied",
    },
    "grant-code": {
        "client_id": "client",
        "expires_in": 1800,
        "user_code": "code",
    },
    "pending-code": {
        "client_id": "client",
        "expires_in": 1800,
        "user_code": "none",
    },
}


class DeviceCodeGrant(_DeviceCodeGrant):
    def query_device_credential(self, device_code):
        data = device_credentials.get(device_code)
        if not data:
            return None

        now = int(time.time())
        data["expires_at"] = now + data["expires_in"]
        data["device_code"] = device_code
        data["scope"] = "profile"
        data["interval"] = 5
        data["verification_uri"] = "https://example.com/activate"
        return DeviceCredentialDict(data)

    def query_user_grant(self, user_code):
        if user_code == "code":
            return db.session.get(User, 1), True
        if user_code == "denied":
            return db.session.get(User, 1), False
        return None

    def should_slow_down(self, credential):
        return False


class DeviceCodeGrantTest(TestCase):
    def create_server(self):
        server = create_authorization_server(self.app)
        server.register_grant(DeviceCodeGrant)
        self.server = server
        return server

    def prepare_data(self, grant_type=DeviceCodeGrant.GRANT_TYPE):
        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="client",
            client_secret="secret",
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["http://localhost/authorized"],
                "scope": "profile",
                "grant_types": [grant_type],
                "token_endpoint_auth_method": "none",
            }
        )
        db.session.add(client)
        db.session.commit()

    def test_invalid_request(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "client_id": "test",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "missing",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

    def test_unauthorized_client(self):
        self.create_server()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "valid-device",
                "client_id": "invalid",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        self.prepare_data(grant_type="password")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "valid-device",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unauthorized_client")

    def test_invalid_client(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "invalid-client",
                "client_id": "invalid",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_expired_token(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "expired-token",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "expired_token")

    def test_denied_by_user(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "denied-code",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "access_denied")

    def test_authorization_pending(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "pending-code",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "authorization_pending")

    def test_get_access_token(self):
        self.create_server()
        self.prepare_data()
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": DeviceCodeGrant.GRANT_TYPE,
                "device_code": "grant-code",
                "client_id": "client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)


class DeviceAuthorizationEndpoint(_DeviceAuthorizationEndpoint):
    def get_verification_uri(self):
        return "https://example.com/activate"

    def save_device_credential(self, client_id, scope, data):
        pass


class DeviceAuthorizationEndpointTest(TestCase):
    def create_server(self):
        server = create_authorization_server(self.app)
        server.register_endpoint(DeviceAuthorizationEndpoint)
        self.server = server

        @self.app.route("/device_authorize", methods=["POST"])
        def device_authorize():
            name = DeviceAuthorizationEndpoint.ENDPOINT_NAME
            return server.create_endpoint_response(name)

        return server

    def test_missing_client_id(self):
        self.create_server()
        rv = self.client.post("/device_authorize", data={"scope": "profile"})
        self.assertEqual(rv.status_code, 401)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_create_authorization_response(self):
        self.create_server()
        client = Client(
            user_id=1,
            client_id="client",
            client_secret="secret",
        )
        db.session.add(client)
        db.session.commit()
        rv = self.client.post(
            "/device_authorize",
            data={
                "client_id": "client",
            },
        )
        self.assertEqual(rv.status_code, 200)
        resp = json.loads(rv.data)
        self.assertIn("device_code", resp)
        self.assertIn("user_code", resp)
        self.assertEqual(resp["verification_uri"], "https://example.com/activate")
        self.assertEqual(
            resp["verification_uri_complete"],
            "https://example.com/activate?user_code=" + resp["user_code"],
        )



import unittest
from collections import OrderedDict

from cryptography.hazmat.primitives.keywrap import InvalidUnwrap

from authlib.common.encoding import json_b64encode
from authlib.common.encoding import json_loads
from authlib.common.encoding import to_bytes
from authlib.common.encoding import urlsafe_b64decode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose import ECKey
from authlib.jose import JsonWebEncryption
from authlib.jose import OKPKey
from authlib.jose.drafts import register_jwe_draft
from authlib.jose.errors import InvalidAlgorithmForMultipleRecipientsMode
from authlib.jose.errors import InvalidEncryptionAlgorithmForECDH1PUWithKeyWrappingError
from authlib.jose.rfc7516.models import JWEHeader

register_jwe_draft(JsonWebEncryption)


class ECDH1PUTest(unittest.TestCase):
    def test_ecdh_1pu_key_agreement_computation_appx_a(self):
        # https://datatracker.ietf.org/doc/html/draft-madden-jose-ecdh-1pu-04#appendix-A
        alice_static_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_static_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }
        alice_ephemeral_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
            "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
            "d": "0_NxaRPUMQoAJt50Gz8YiTr8gRTwyEaCumd-MToTmIo",
        }

        headers = {
            "alg": "ECDH-1PU",
            "enc": "A256GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
            "epk": {
                "kty": "EC",
                "crv": "P-256",
                "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
                "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
            },
        }

        alg = JsonWebEncryption.ALG_REGISTRY["ECDH-1PU"]
        enc = JsonWebEncryption.ENC_REGISTRY["A256GCM"]

        alice_static_key = alg.prepare_key(alice_static_key)
        bob_static_key = alg.prepare_key(bob_static_key)
        alice_ephemeral_key = alg.prepare_key(alice_ephemeral_key)

        alice_static_pubkey = alice_static_key.get_op_key("wrapKey")
        bob_static_pubkey = bob_static_key.get_op_key("wrapKey")
        alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")

        # Derived key computation at Alice

        # Step-by-step methods verification
        _shared_key_e_at_alice = alice_ephemeral_key.exchange_shared_key(
            bob_static_pubkey
        )
        self.assertEqual(
            _shared_key_e_at_alice,
            b"\x9e\x56\xd9\x1d\x81\x71\x35\xd3\x72\x83\x42\x83\xbf\x84\x26\x9c"
            + b"\xfb\x31\x6e\xa3\xda\x80\x6a\x48\xf6\xda\xa7\x79\x8c\xfe\x90\xc4",
        )

        _shared_key_s_at_alice = alice_static_key.exchange_shared_key(bob_static_pubkey)
        self.assertEqual(
            _shared_key_s_at_alice,
            b"\xe3\xca\x34\x74\x38\x4c\x9f\x62\xb3\x0b\xfd\x4c\x68\x8b\x3e\x7d"
            + b"\x41\x10\xa1\xb4\xba\xdc\x3c\xc5\x4e\xf7\xb8\x12\x41\xef\xd5\x0d",
        )

        _shared_key_at_alice = alg.compute_shared_key(
            _shared_key_e_at_alice, _shared_key_s_at_alice
        )
        self.assertEqual(
            _shared_key_at_alice,
            b"\x9e\x56\xd9\x1d\x81\x71\x35\xd3\x72\x83\x42\x83\xbf\x84\x26\x9c"
            + b"\xfb\x31\x6e\xa3\xda\x80\x6a\x48\xf6\xda\xa7\x79\x8c\xfe\x90\xc4"
            + b"\xe3\xca\x34\x74\x38\x4c\x9f\x62\xb3\x0b\xfd\x4c\x68\x8b\x3e\x7d"
            + b"\x41\x10\xa1\xb4\xba\xdc\x3c\xc5\x4e\xf7\xb8\x12\x41\xef\xd5\x0d",
        )

        _fixed_info_at_alice = alg.compute_fixed_info(headers, enc.key_size, None)
        self.assertEqual(
            _fixed_info_at_alice,
            b"\x00\x00\x00\x07\x41\x32\x35\x36\x47\x43\x4d\x00\x00\x00\x05\x41"
            + b"\x6c\x69\x63\x65\x00\x00\x00\x03\x42\x6f\x62\x00\x00\x01\x00",
        )

        _dk_at_alice = alg.compute_derived_key(
            _shared_key_at_alice, _fixed_info_at_alice, enc.key_size
        )
        self.assertEqual(
            _dk_at_alice,
            b"\x6c\xaf\x13\x72\x3d\x14\x85\x0a\xd4\xb4\x2c\xd6\xdd\xe9\x35\xbf"
            + b"\xfd\x2f\xff\x00\xa9\xba\x70\xde\x05\xc2\x03\xa5\xe1\x72\x2c\xa7",
        )
        self.assertEqual(
            urlsafe_b64encode(_dk_at_alice),
            b"bK8Tcj0UhQrUtCzW3ek1v_0v_wCpunDeBcIDpeFyLKc",
        )

        # All-in-one method verification
        dk_at_alice = alg.deliver_at_sender(
            alice_static_key,
            alice_ephemeral_key,
            bob_static_pubkey,
            headers,
            enc.key_size,
            None,
        )
        self.assertEqual(
            urlsafe_b64encode(dk_at_alice),
            b"bK8Tcj0UhQrUtCzW3ek1v_0v_wCpunDeBcIDpeFyLKc",
        )

        # Derived key computation at Bob

        # Step-by-step methods verification
        _shared_key_e_at_bob = bob_static_key.exchange_shared_key(
            alice_ephemeral_pubkey
        )
        self.assertEqual(_shared_key_e_at_bob, _shared_key_e_at_alice)

        _shared_key_s_at_bob = bob_static_key.exchange_shared_key(alice_static_pubkey)
        self.assertEqual(_shared_key_s_at_bob, _shared_key_s_at_alice)

        _shared_key_at_bob = alg.compute_shared_key(
            _shared_key_e_at_bob, _shared_key_s_at_bob
        )
        self.assertEqual(_shared_key_at_bob, _shared_key_at_alice)

        _fixed_info_at_bob = alg.compute_fixed_info(headers, enc.key_size, None)
        self.assertEqual(_fixed_info_at_bob, _fixed_info_at_alice)

        _dk_at_bob = alg.compute_derived_key(
            _shared_key_at_bob, _fixed_info_at_bob, enc.key_size
        )
        self.assertEqual(_dk_at_bob, _dk_at_alice)

        # All-in-one method verification
        dk_at_bob = alg.deliver_at_recipient(
            bob_static_key,
            alice_static_pubkey,
            alice_ephemeral_pubkey,
            headers,
            enc.key_size,
            None,
        )
        self.assertEqual(dk_at_bob, dk_at_alice)

    def test_ecdh_1pu_key_agreement_computation_appx_b(self):
        # https://datatracker.ietf.org/doc/html/draft-madden-jose-ecdh-1pu-04#appendix-B
        alice_static_key = {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
        bob_static_key = {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
        charlie_static_key = {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
        alice_ephemeral_key = {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
            "d": "x8EVZH4Fwk673_mUujnliJoSrLz0zYzzCWp5GUX2fc8",
        }

        protected = OrderedDict(
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": OrderedDict(
                    {
                        "kty": "OKP",
                        "crv": "X25519",
                        "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                    }
                ),
            }
        )

        cek = (
            b"\xff\xfe\xfd\xfc\xfb\xfa\xf9\xf8\xf7\xf6\xf5\xf4\xf3\xf2\xf1\xf0"
            b"\xef\xee\xed\xec\xeb\xea\xe9\xe8\xe7\xe6\xe5\xe4\xe3\xe2\xe1\xe0"
            b"\xdf\xde\xdd\xdc\xdb\xda\xd9\xd8\xd7\xd6\xd5\xd4\xd3\xd2\xd1\xd0"
            b"\xcf\xce\xcd\xcc\xcb\xca\xc9\xc8\xc7\xc6\xc5\xc4\xc3\xc2\xc1\xc0"
        )

        iv = b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"

        payload = b"Three is a magic number."

        alg = JsonWebEncryption.ALG_REGISTRY["ECDH-1PU+A128KW"]
        enc = JsonWebEncryption.ENC_REGISTRY["A256CBC-HS512"]

        alice_static_key = OKPKey.import_key(alice_static_key)
        bob_static_key = OKPKey.import_key(bob_static_key)
        charlie_static_key = OKPKey.import_key(charlie_static_key)
        alice_ephemeral_key = OKPKey.import_key(alice_ephemeral_key)

        alice_static_pubkey = alice_static_key.get_op_key("wrapKey")
        bob_static_pubkey = bob_static_key.get_op_key("wrapKey")
        charlie_static_pubkey = charlie_static_key.get_op_key("wrapKey")
        alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")

        protected_segment = json_b64encode(protected)
        aad = to_bytes(protected_segment, "ascii")

        ciphertext, tag = enc.encrypt(payload, aad, iv, cek)
        self.assertEqual(
            urlsafe_b64encode(ciphertext),
            b"Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
        )
        self.assertEqual(
            urlsafe_b64encode(tag), b"HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
        )

        # Derived key computation at Alice for Bob

        # Step-by-step methods verification
        _shared_key_e_at_alice_for_bob = alice_ephemeral_key.exchange_shared_key(
            bob_static_pubkey
        )
        self.assertEqual(
            _shared_key_e_at_alice_for_bob,
            b"\x32\x81\x08\x96\xe0\xfe\x4d\x57\x0e\xd1\xac\xfc\xed\xf6\x71\x17"
            + b"\xdc\x19\x4e\xd5\xda\xac\x21\xd8\xff\x7a\xf3\x24\x46\x94\x89\x7f",
        )

        _shared_key_s_at_alice_for_bob = alice_static_key.exchange_shared_key(
            bob_static_pubkey
        )
        self.assertEqual(
            _shared_key_s_at_alice_for_bob,
            b"\x21\x57\x61\x2c\x90\x48\xed\xfa\xe7\x7c\xb2\xe4\x23\x71\x40\x60"
            + b"\x59\x67\xc0\x5c\x7f\x77\xa4\x8e\xea\xf2\xcf\x29\xa5\x73\x7c\x4a",
        )

        _shared_key_at_alice_for_bob = alg.compute_shared_key(
            _shared_key_e_at_alice_for_bob, _shared_key_s_at_alice_for_bob
        )
        self.assertEqual(
            _shared_key_at_alice_for_bob,
            b"\x32\x81\x08\x96\xe0\xfe\x4d\x57\x0e\xd1\xac\xfc\xed\xf6\x71\x17"
            + b"\xdc\x19\x4e\xd5\xda\xac\x21\xd8\xff\x7a\xf3\x24\x46\x94\x89\x7f"
            + b"\x21\x57\x61\x2c\x90\x48\xed\xfa\xe7\x7c\xb2\xe4\x23\x71\x40\x60"
            + b"\x59\x67\xc0\x5c\x7f\x77\xa4\x8e\xea\xf2\xcf\x29\xa5\x73\x7c\x4a",
        )

        _fixed_info_at_alice_for_bob = alg.compute_fixed_info(
            protected, alg.key_size, tag
        )
        self.assertEqual(
            _fixed_info_at_alice_for_bob,
            b"\x00\x00\x00\x0f\x45\x43\x44\x48\x2d\x31\x50\x55\x2b\x41\x31\x32"
            + b"\x38\x4b\x57\x00\x00\x00\x05\x41\x6c\x69\x63\x65\x00\x00\x00\x0f"
            + b"\x42\x6f\x62\x20\x61\x6e\x64\x20\x43\x68\x61\x72\x6c\x69\x65\x00"
            + b"\x00\x00\x80\x00\x00\x00\x20\x1c\xb6\xf8\x7d\x39\x66\xf2\xca\x46"
            + b"\x9a\x28\xf7\x47\x23\xac\xda\x02\x78\x0e\x91\xcc\xe2\x18\x55\x47"
            + b"\x07\x45\xfe\x11\x9b\xdd\x64",
        )

        _dk_at_alice_for_bob = alg.compute_derived_key(
            _shared_key_at_alice_for_bob, _fixed_info_at_alice_for_bob, alg.key_size
        )
        self.assertEqual(
            _dk_at_alice_for_bob,
            b"\xdf\x4c\x37\xa0\x66\x83\x06\xa1\x1e\x3d\x6b\x00\x74\xb5\xd8\xdf",
        )

        # All-in-one method verification
        dk_at_alice_for_bob = alg.deliver_at_sender(
            alice_static_key,
            alice_ephemeral_key,
            bob_static_pubkey,
            protected,
            alg.key_size,
            tag,
        )
        self.assertEqual(
            dk_at_alice_for_bob,
            b"\xdf\x4c\x37\xa0\x66\x83\x06\xa1\x1e\x3d\x6b\x00\x74\xb5\xd8\xdf",
        )

        kek_at_alice_for_bob = alg.aeskw.prepare_key(dk_at_alice_for_bob)
        wrapped_for_bob = alg.aeskw.wrap_cek(cek, kek_at_alice_for_bob)
        ek_for_bob = wrapped_for_bob["ek"]
        self.assertEqual(
            urlsafe_b64encode(ek_for_bob),
            b"pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN",
        )

        # Derived key computation at Alice for Charlie

        # Step-by-step methods verification
        _shared_key_e_at_alice_for_charlie = alice_ephemeral_key.exchange_shared_key(
            charlie_static_pubkey
        )
        self.assertEqual(
            _shared_key_e_at_alice_for_charlie,
            b"\x89\xdc\xfe\x4c\x37\xc1\xdc\x02\x71\xf3\x46\xb5\xb3\xb1\x9c\x3b"
            + b"\x70\x5c\xa2\xa7\x2f\x9a\x23\x77\x85\xc3\x44\x06\xfc\xb7\x5f\x10",
        )

        _shared_key_s_at_alice_for_charlie = alice_static_key.exchange_shared_key(
            charlie_static_pubkey
        )
        self.assertEqual(
            _shared_key_s_at_alice_for_charlie,
            b"\x78\xfe\x63\xfc\x66\x1c\xf8\xd1\x8f\x92\xa8\x42\x2a\x64\x18\xe4"
            + b"\xed\x5e\x20\xa9\x16\x81\x85\xfd\xee\xdc\xa1\xc3\xd8\xe6\xa6\x1c",
        )

        _shared_key_at_alice_for_charlie = alg.compute_shared_key(
            _shared_key_e_at_alice_for_charlie, _shared_key_s_at_alice_for_charlie
        )
        self.assertEqual(
            _shared_key_at_alice_for_charlie,
            b"\x89\xdc\xfe\x4c\x37\xc1\xdc\x02\x71\xf3\x46\xb5\xb3\xb1\x9c\x3b"
            + b"\x70\x5c\xa2\xa7\x2f\x9a\x23\x77\x85\xc3\x44\x06\xfc\xb7\x5f\x10"
            + b"\x78\xfe\x63\xfc\x66\x1c\xf8\xd1\x8f\x92\xa8\x42\x2a\x64\x18\xe4"
            + b"\xed\x5e\x20\xa9\x16\x81\x85\xfd\xee\xdc\xa1\xc3\xd8\xe6\xa6\x1c",
        )

        _fixed_info_at_alice_for_charlie = alg.compute_fixed_info(
            protected, alg.key_size, tag
        )
        self.assertEqual(_fixed_info_at_alice_for_charlie, _fixed_info_at_alice_for_bob)

        _dk_at_alice_for_charlie = alg.compute_derived_key(
            _shared_key_at_alice_for_charlie,
            _fixed_info_at_alice_for_charlie,
            alg.key_size,
        )
        self.assertEqual(
            _dk_at_alice_for_charlie,
            b"\x57\xd8\x12\x6f\x1b\x7e\xc4\xcc\xb0\x58\x4d\xac\x03\xcb\x27\xcc",
        )

        # All-in-one method verification
        dk_at_alice_for_charlie = alg.deliver_at_sender(
            alice_static_key,
            alice_ephemeral_key,
            charlie_static_pubkey,
            protected,
            alg.key_size,
            tag,
        )
        self.assertEqual(
            dk_at_alice_for_charlie,
            b"\x57\xd8\x12\x6f\x1b\x7e\xc4\xcc\xb0\x58\x4d\xac\x03\xcb\x27\xcc",
        )

        kek_at_alice_for_charlie = alg.aeskw.prepare_key(dk_at_alice_for_charlie)
        wrapped_for_charlie = alg.aeskw.wrap_cek(cek, kek_at_alice_for_charlie)
        ek_for_charlie = wrapped_for_charlie["ek"]
        self.assertEqual(
            urlsafe_b64encode(ek_for_charlie),
            b"56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE",
        )

        # Derived key computation at Bob for Alice

        # Step-by-step methods verification
        _shared_key_e_at_bob_for_alice = bob_static_key.exchange_shared_key(
            alice_ephemeral_pubkey
        )
        self.assertEqual(_shared_key_e_at_bob_for_alice, _shared_key_e_at_alice_for_bob)

        _shared_key_s_at_bob_for_alice = bob_static_key.exchange_shared_key(
            alice_static_pubkey
        )
        self.assertEqual(_shared_key_s_at_bob_for_alice, _shared_key_s_at_alice_for_bob)

        _shared_key_at_bob_for_alice = alg.compute_shared_key(
            _shared_key_e_at_bob_for_alice, _shared_key_s_at_bob_for_alice
        )
        self.assertEqual(_shared_key_at_bob_for_alice, _shared_key_at_alice_for_bob)

        _fixed_info_at_bob_for_alice = alg.compute_fixed_info(
            protected, alg.key_size, tag
        )
        self.assertEqual(_fixed_info_at_bob_for_alice, _fixed_info_at_alice_for_bob)

        _dk_at_bob_for_alice = alg.compute_derived_key(
            _shared_key_at_bob_for_alice, _fixed_info_at_bob_for_alice, alg.key_size
        )
        self.assertEqual(_dk_at_bob_for_alice, _dk_at_alice_for_bob)

        # All-in-one method verification
        dk_at_bob_for_alice = alg.deliver_at_recipient(
            bob_static_key,
            alice_static_pubkey,
            alice_ephemeral_pubkey,
            protected,
            alg.key_size,
            tag,
        )
        self.assertEqual(dk_at_bob_for_alice, dk_at_alice_for_bob)

        kek_at_bob_for_alice = alg.aeskw.prepare_key(dk_at_bob_for_alice)
        cek_unwrapped_by_bob = alg.aeskw.unwrap(
            enc, ek_for_bob, protected, kek_at_bob_for_alice
        )
        self.assertEqual(cek_unwrapped_by_bob, cek)

        payload_decrypted_by_bob = enc.decrypt(
            ciphertext, aad, iv, tag, cek_unwrapped_by_bob
        )
        self.assertEqual(payload_decrypted_by_bob, payload)

        # Derived key computation at Charlie for Alice

        # Step-by-step methods verification
        _shared_key_e_at_charlie_for_alice = charlie_static_key.exchange_shared_key(
            alice_ephemeral_pubkey
        )
        self.assertEqual(
            _shared_key_e_at_charlie_for_alice, _shared_key_e_at_alice_for_charlie
        )

        _shared_key_s_at_charlie_for_alice = charlie_static_key.exchange_shared_key(
            alice_static_pubkey
        )
        self.assertEqual(
            _shared_key_s_at_charlie_for_alice, _shared_key_s_at_alice_for_charlie
        )

        _shared_key_at_charlie_for_alice = alg.compute_shared_key(
            _shared_key_e_at_charlie_for_alice, _shared_key_s_at_charlie_for_alice
        )
        self.assertEqual(
            _shared_key_at_charlie_for_alice, _shared_key_at_alice_for_charlie
        )

        _fixed_info_at_charlie_for_alice = alg.compute_fixed_info(
            protected, alg.key_size, tag
        )
        self.assertEqual(
            _fixed_info_at_charlie_for_alice, _fixed_info_at_alice_for_charlie
        )

        _dk_at_charlie_for_alice = alg.compute_derived_key(
            _shared_key_at_charlie_for_alice,
            _fixed_info_at_charlie_for_alice,
            alg.key_size,
        )
        self.assertEqual(_dk_at_charlie_for_alice, _dk_at_alice_for_charlie)

        # All-in-one method verification
        dk_at_charlie_for_alice = alg.deliver_at_recipient(
            charlie_static_key,
            alice_static_pubkey,
            alice_ephemeral_pubkey,
            protected,
            alg.key_size,
            tag,
        )
        self.assertEqual(dk_at_charlie_for_alice, dk_at_alice_for_charlie)

        kek_at_charlie_for_alice = alg.aeskw.prepare_key(dk_at_charlie_for_alice)
        cek_unwrapped_by_charlie = alg.aeskw.unwrap(
            enc, ek_for_charlie, protected, kek_at_charlie_for_alice
        )
        self.assertEqual(cek_unwrapped_by_charlie, cek)

        payload_decrypted_by_charlie = enc.decrypt(
            ciphertext, aad, iv, tag, cek_unwrapped_by_charlie
        )
        self.assertEqual(payload_decrypted_by_charlie, payload)

    def test_ecdh_1pu_jwe_in_direct_key_agreement_mode(self):
        jwe = JsonWebEncryption()
        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": "ECDH-1PU", "enc": enc}
            data = jwe.serialize_compact(
                protected, b"hello", bob_key, sender_key=alice_key
            )
            rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
            self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_jwe_json_serialization_single_recipient_in_direct_key_agreement_mode(
        self,
    ):
        jwe = JsonWebEncryption()
        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-1PU", "enc": "A128GCM"}
        header_obj = {"protected": protected}
        data = jwe.serialize_json(header_obj, b"hello", bob_key, sender_key=alice_key)
        rv = jwe.deserialize_json(data, bob_key, sender_key=alice_key)
        self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_jwe_in_key_agreement_with_key_wrapping_mode(self):
        jwe = JsonWebEncryption()
        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        for alg in [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]:
            for enc in [
                "A128CBC-HS256",
                "A192CBC-HS384",
                "A256CBC-HS512",
            ]:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(
                    protected, b"hello", bob_key, sender_key=alice_key
                )
                rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
                self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_jwe_with_compact_serialization_ignores_kid_provided_separately_on_decryption(
        self,
    ):
        jwe = JsonWebEncryption()

        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }

        bob_kid = "Bob's key"
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        for alg in [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]:
            for enc in [
                "A128CBC-HS256",
                "A192CBC-HS384",
                "A256CBC-HS512",
            ]:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(
                    protected, b"hello", bob_key, sender_key=alice_key
                )
                rv = jwe.deserialize_compact(
                    data, (bob_kid, bob_key), sender_key=alice_key
                )
                self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_jwe_with_okp_keys_in_direct_key_agreement_mode(self):
        jwe = JsonWebEncryption()
        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": "ECDH-1PU", "enc": enc}
            data = jwe.serialize_compact(
                protected, b"hello", bob_key, sender_key=alice_key
            )
            rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
            self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_jwe_with_okp_keys_in_key_agreement_with_key_wrapping_mode(self):
        jwe = JsonWebEncryption()
        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        for alg in [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]:
            for enc in [
                "A128CBC-HS256",
                "A192CBC-HS384",
                "A256CBC-HS512",
            ]:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(
                    protected, b"hello", bob_key, sender_key=alice_key
                )
                rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
                self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_1pu_encryption_with_json_serialization(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {"header": {"kid": "bob-key-2"}},
            {"header": {"kid": "2021-05-06"}},
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(
            header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
        )

        self.assertEqual(
            data.keys(),
            {
                "protected",
                "unprotected",
                "recipients",
                "aad",
                "iv",
                "ciphertext",
                "tag",
            },
        )

        decoded_protected = json_loads(
            urlsafe_b64decode(to_bytes(data["protected"])).decode("utf-8")
        )
        self.assertEqual(decoded_protected.keys(), protected.keys() | {"epk"})
        self.assertEqual(
            {k: decoded_protected[k] for k in decoded_protected.keys() - {"epk"}},
            protected,
        )

        self.assertEqual(data["unprotected"], unprotected)

        self.assertEqual(len(data["recipients"]), len(recipients))
        for i in range(len(data["recipients"])):
            self.assertEqual(data["recipients"][i].keys(), {"header", "encrypted_key"})
            self.assertEqual(data["recipients"][i]["header"], recipients[i]["header"])

        self.assertEqual(urlsafe_b64decode(to_bytes(data["aad"])), jwe_aad)

        iv = urlsafe_b64decode(to_bytes(data["iv"]))
        ciphertext = urlsafe_b64decode(to_bytes(data["ciphertext"]))
        tag = urlsafe_b64decode(to_bytes(data["tag"]))

        alg = JsonWebEncryption.ALG_REGISTRY[protected["alg"]]
        enc = JsonWebEncryption.ENC_REGISTRY[protected["enc"]]

        aad = to_bytes(data["protected"]) + b"." + to_bytes(data["aad"])
        aad = to_bytes(aad, "ascii")

        ek_for_bob = urlsafe_b64decode(to_bytes(data["recipients"][0]["encrypted_key"]))
        header_for_bob = JWEHeader(
            decoded_protected, data["unprotected"], data["recipients"][0]["header"]
        )
        cek_at_bob = alg.unwrap(
            enc, ek_for_bob, header_for_bob, bob_key, sender_key=alice_key, tag=tag
        )
        payload_at_bob = enc.decrypt(ciphertext, aad, iv, tag, cek_at_bob)

        self.assertEqual(payload_at_bob, payload)

        ek_for_charlie = urlsafe_b64decode(
            to_bytes(data["recipients"][1]["encrypted_key"])
        )
        header_for_charlie = JWEHeader(
            decoded_protected, data["unprotected"], data["recipients"][1]["header"]
        )
        cek_at_charlie = alg.unwrap(
            enc,
            ek_for_charlie,
            header_for_charlie,
            charlie_key,
            sender_key=alice_key,
            tag=tag,
        )
        payload_at_charlie = enc.decrypt(ciphertext, aad, iv, tag, cek_at_charlie)

        self.assertEqual(cek_at_charlie, cek_at_bob)
        self.assertEqual(payload_at_charlie, payload)

    def test_ecdh_1pu_decryption_with_json_serialization(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        data = {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
            + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
            + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
            + "RnFVQUZhMzlkeUJjIn19",
            "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"kid": "bob-key-2"},
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                    + "eU1cSl55cQ0hGezJu2N9IY0QN",
                },
                {
                    "header": {"kid": "2021-05-06"},
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                    + "fe4z3PQ2YH2afvjQ28aiCTWFE",
                },
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
        }

        rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

        self.assertEqual(rv_at_bob.keys(), {"header", "payload"})

        self.assertEqual(
            rv_at_bob["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv_at_bob["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv_at_bob["header"]["unprotected"],
            {"jku": "https://alice.example.com/keys.jwks"},
        )

        self.assertEqual(
            rv_at_bob["header"]["recipients"],
            [{"header": {"kid": "bob-key-2"}}, {"header": {"kid": "2021-05-06"}}],
        )

        self.assertEqual(rv_at_bob["payload"], b"Three is a magic number.")

        rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

        self.assertEqual(rv_at_charlie.keys(), {"header", "payload"})

        self.assertEqual(
            rv_at_charlie["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv_at_charlie["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv_at_charlie["header"]["unprotected"],
            {"jku": "https://alice.example.com/keys.jwks"},
        )

        self.assertEqual(
            rv_at_charlie["header"]["recipients"],
            [{"header": {"kid": "bob-key-2"}}, {"header": {"kid": "2021-05-06"}}],
        )

        self.assertEqual(rv_at_charlie["payload"], b"Three is a magic number.")

    def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_not_specified(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {"header": {"kid": "bob-key-2"}},
            {"header": {"kid": "2021-05-06"}},
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(
            header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
        )

        rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

        self.assertEqual(
            rv_at_bob["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_bob["header"]["protected"][k]
                for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_bob["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_bob["header"]["recipients"], recipients)
        self.assertEqual(rv_at_bob["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_bob["payload"], payload)

        rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

        self.assertEqual(
            rv_at_charlie["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_charlie["header"]["protected"][k]
                for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_charlie["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_charlie["header"]["recipients"], recipients)
        self.assertEqual(rv_at_charlie["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_charlie["payload"], payload)

    def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_specified(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "alice-key",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "bob-key-2",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "2021-05-06",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {"header": {"kid": "bob-key-2"}},
            {"header": {"kid": "2021-05-06"}},
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(
            header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
        )

        rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

        self.assertEqual(
            rv_at_bob["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_bob["header"]["protected"][k]
                for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_bob["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_bob["header"]["recipients"], recipients)
        self.assertEqual(rv_at_bob["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_bob["payload"], payload)

        rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

        self.assertEqual(
            rv_at_charlie["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_charlie["header"]["protected"][k]
                for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_charlie["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_charlie["header"]["recipients"], recipients)
        self.assertEqual(rv_at_charlie["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_charlie["payload"], payload)

    def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_provided_separately_on_decryption(
        self,
    ):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "WjKgJV7VRw3hmgU6--4v15c0Aewbcvat1BsRFTIqa5Q",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )

        bob_kid = "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )

        charlie_kid = "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {
                "header": {
                    "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
                }
            },
            {
                "header": {
                    "kid": "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
                }
            },
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(
            header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
        )

        rv_at_bob = jwe.deserialize_json(data, (bob_kid, bob_key), sender_key=alice_key)

        self.assertEqual(
            rv_at_bob["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_bob["header"]["protected"][k]
                for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_bob["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_bob["header"]["recipients"], recipients)
        self.assertEqual(rv_at_bob["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_bob["payload"], payload)

        rv_at_charlie = jwe.deserialize_json(
            data, (charlie_kid, charlie_key), sender_key=alice_key
        )

        self.assertEqual(
            rv_at_charlie["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_charlie["header"]["protected"][k]
                for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_charlie["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_charlie["header"]["recipients"], recipients)
        self.assertEqual(rv_at_charlie["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_charlie["payload"], payload)

    def test_ecdh_1pu_jwe_with_json_serialization_for_single_recipient(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [{"header": {"kid": "bob-key-2"}}]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, bob_key, sender_key=alice_key)

        rv = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

        self.assertEqual(rv["header"]["protected"].keys(), protected.keys() | {"epk"})
        self.assertEqual(
            {
                k: rv["header"]["protected"][k]
                for k in rv["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv["header"]["unprotected"], unprotected)
        self.assertEqual(rv["header"]["recipients"], recipients)
        self.assertEqual(rv["header"]["aad"], jwe_aad)
        self.assertEqual(rv["payload"], payload)

    def test_ecdh_1pu_encryption_fails_json_serialization_multiple_recipients_in_direct_key_agreement_mode(
        self,
    ):
        jwe = JsonWebEncryption()
        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)
        charlie_key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-1PU", "enc": "A128GCM"}
        header_obj = {"protected": protected}
        self.assertRaises(
            InvalidAlgorithmForMultipleRecipientsMode,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_fails_if_not_aes_cbc_hmac_sha2_enc_is_used_with_kw(
        self,
    ):
        jwe = JsonWebEncryption()
        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        }

        for alg in [
            "ECDH-1PU+A128KW",
            "ECDH-1PU+A192KW",
            "ECDH-1PU+A256KW",
        ]:
            for enc in [
                "A128GCM",
                "A192GCM",
                "A256GCM",
            ]:
                protected = {"alg": alg, "enc": enc}
                self.assertRaises(
                    InvalidEncryptionAlgorithmForECDH1PUWithKeyWrappingError,
                    jwe.serialize_compact,
                    protected,
                    b"hello",
                    bob_key,
                    sender_key=alice_key,
                )

    def test_ecdh_1pu_encryption_with_public_sender_key_fails(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }
        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_ecdh_1pu_decryption_with_public_recipient_key_fails(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        }
        data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
        self.assertRaises(
            ValueError, jwe.deserialize_compact, data, bob_key, sender_key=alice_key
        )

    def test_ecdh_1pu_encryption_fails_if_key_types_are_different(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = ECKey.generate_key("P-256", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=False)
        self.assertRaises(
            Exception,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = ECKey.generate_key("P-256", is_private=False)
        self.assertRaises(
            Exception,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_fails_if_keys_curves_are_different(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = ECKey.generate_key("P-256", is_private=True)
        bob_key = ECKey.generate_key("secp256k1", is_private=False)
        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = ECKey.generate_key("P-384", is_private=True)
        bob_key = ECKey.generate_key("P-521", is_private=False)
        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X448", is_private=False)
        self.assertRaises(
            TypeError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_fails_if_key_points_are_not_actually_on_same_curve(
        self,
    ):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "aDHtGkIYyhR5geqfMaFL0T9cG4JEMI8nyMFJA7gRUDs",
            "y": "AjGN5_f-aCt4vYg74my6n1ALIq746nlc_httIgcBSYY",
            "d": "Sim3EIzXsWaWu9QW8yKVHwxBM5CTlnrVU_Eq-y_KRQA",
        }  # the point is indeed on P-256 curve
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "5ZFnZbs_BtLBIZxwt5hS7SBDtI2a-dJ871dJ8ZnxZ6c",
            "y": "K0srqSkbo1Yeckr0YoQA8r_rOz0ZUStiv3mc1qn46pg",
        }  # the point is not on P-256 curve but is actually on secp256k1 curve

        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = {
            "kty": "EC",
            "crv": "P-521",
            "x": "1JDMOjnMgASo01PVHRcyCDtE6CLgKuwXLXLbdLGxpdubLuHYBa0KAepyimnxCWsX",
            "y": "w7BSC8Xb3XgMMfE7IFCJpoOmx1Sf3T3_3OZ4CrF6_iCFAw4VOdFYR42OnbKMFG--",
            "d": "lCkpFBaVwHzfHtkJEV3PzxefObOPnMgUjNZSLryqC5AkERgXT3-DZLEi6eBzq5gk",
        }  # the point is not on P-521 curve but is actually on P-384 curve
        bob_key = {
            "kty": "EC",
            "crv": "P-521",
            "x": "Cd6rinJdgS4WJj6iaNyXiVhpMbhZLmPykmrnFhIad04B3ulf5pURb5v9mx21c_Cv8Q1RBOptwleLg5Qjq2J1qa4",
            "y": "hXo9p1EjW6W4opAQdmfNgyxztkNxYwn9L4FVTLX51KNEsW0aqueLm96adRmf0HoGIbNhIdcIlXOKlRUHqgunDkM",
        }  # the point is indeed on P-521 curve

        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "TAB1oIsjPob3guKwTEeQsAsupSRPdXdxHhnV8JrVJTA",
                "d": "kO2LzPr4vLg_Hn-7_MDq66hJZgvTIkzDG4p6nCsgNHk",
            }
        )  # the point is indeed on X25519 curve
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "lVHcPx4R9bExaoxXZY9tAq7SNW9pJKCoVQxURLtkAs3Dg5ZRxcjhf0JUyg2lod5OGDptJ7wowwY",
            }
        )  # the point is not on X25519 curve but is actually on X448 curve

        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X448",
                "x": "TAB1oIsjPob3guKwTEeQsAsupSRPdXdxHhnV8JrVJTA",
                "d": "kO2LzPr4vLg_Hn-7_MDq66hJZgvTIkzDG4p6nCsgNHk",
            }
        )  # the point is not on X448 curve but is actually on X25519 curve
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X448",
                "x": "lVHcPx4R9bExaoxXZY9tAq7SNW9pJKCoVQxURLtkAs3Dg5ZRxcjhf0JUyg2lod5OGDptJ7wowwY",
            }
        )  # the point is indeed on X448 curve

        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_fails_if_keys_curve_is_inappropriate(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

        alice_key = OKPKey.generate_key(
            "Ed25519", is_private=True
        )  # use Ed25519 instead of X25519
        bob_key = OKPKey.generate_key(
            "Ed25519", is_private=False
        )  # use Ed25519 instead of X25519
        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_key_types_are_different(
        self,
    ):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        alice_key = ECKey.generate_key("P-256", is_private=True)
        bob_key = ECKey.generate_key("P-256", is_private=False)
        charlie_key = OKPKey.generate_key("X25519", is_private=False)

        self.assertRaises(
            Exception,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_keys_curves_are_different(
        self,
    ):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X448", is_private=False)
        charlie_key = OKPKey.generate_key("X25519", is_private=False)

        self.assertRaises(
            TypeError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_key_points_are_not_actually_on_same_curve(
        self,
    ):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "aDHtGkIYyhR5geqfMaFL0T9cG4JEMI8nyMFJA7gRUDs",
            "y": "AjGN5_f-aCt4vYg74my6n1ALIq746nlc_httIgcBSYY",
            "d": "Sim3EIzXsWaWu9QW8yKVHwxBM5CTlnrVU_Eq-y_KRQA",
        }  # the point is indeed on P-256 curve
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "HgF88mm6yw4gjG7yG6Sqz66pHnpZcyx7c842BQghYuc",
            "y": "KZ1ywvTOYnpNb4Gepa5eSgfEOb5gj5hCaCFIrTFuI2o",
        }  # the point is indeed on P-256 curve
        charlie_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "5ZFnZbs_BtLBIZxwt5hS7SBDtI2a-dJ871dJ8ZnxZ6c",
            "y": "K0srqSkbo1Yeckr0YoQA8r_rOz0ZUStiv3mc1qn46pg",
        }  # the point is not on P-256 curve but is actually on secp256k1 curve

        self.assertRaises(
            ValueError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )

    def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_keys_curve_is_inappropriate(
        self,
    ):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        alice_key = OKPKey.generate_key(
            "Ed25519", is_private=True
        )  # use Ed25519 instead of X25519
        bob_key = OKPKey.generate_key(
            "Ed25519", is_private=False
        )  # use Ed25519 instead of X25519
        charlie_key = OKPKey.generate_key(
            "Ed25519", is_private=False
        )  # use Ed25519 instead of X25519

        self.assertRaises(
            ValueError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )

    def test_ecdh_1pu_decryption_fails_if_key_matches_to_no_recipient(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [{"header": {"kid": "bob-key-2"}}]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, bob_key, sender_key=alice_key)

        self.assertRaises(
            InvalidUnwrap, jwe.deserialize_json, data, charlie_key, sender_key=alice_key
        )



import unittest

from authlib.common.encoding import base64_to_int
from authlib.common.encoding import json_dumps
from authlib.jose import ECKey
from authlib.jose import JsonWebKey
from authlib.jose import KeySet
from authlib.jose import OctKey
from authlib.jose import OKPKey
from authlib.jose import RSAKey
from tests.util import read_file_path


class BaseTest(unittest.TestCase):
    def assertBase64IntEqual(self, x, y):
        self.assertEqual(base64_to_int(x), base64_to_int(y))


class OctKeyTest(BaseTest):
    def test_import_oct_key(self):
        # https://tools.ietf.org/html/rfc7520#section-3.5
        obj = {
            "kty": "oct",
            "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037",
            "use": "sig",
            "alg": "HS256",
            "k": "hJtXIZ2uSN5kbQfbtTNWbpdmhkV8FJG-Onbc6mxCcYg",
        }
        key = OctKey.import_key(obj)
        new_obj = key.as_dict()
        self.assertEqual(obj["k"], new_obj["k"])
        self.assertIn("use", new_obj)

    def test_invalid_oct_key(self):
        self.assertRaises(ValueError, OctKey.import_key, {})

    def test_generate_oct_key(self):
        self.assertRaises(ValueError, OctKey.generate_key, 251)

        with self.assertRaises(ValueError) as cm:
            OctKey.generate_key(is_private=False)

        self.assertEqual(str(cm.exception), "oct key can not be generated as public")

        key = OctKey.generate_key()
        self.assertIn("kid", key.as_dict())
        self.assertNotIn("use", key.as_dict())

        key2 = OctKey.import_key(key, {"use": "sig"})
        self.assertIn("use", key2.as_dict())


class RSAKeyTest(BaseTest):
    def test_import_ssh_pem(self):
        raw = read_file_path("ssh_public.pem")
        key = RSAKey.import_key(raw)
        obj = key.as_dict()
        self.assertEqual(obj["kty"], "RSA")

    def test_rsa_public_key(self):
        # https://tools.ietf.org/html/rfc7520#section-3.3
        obj = read_file_path("jwk_public.json")
        key = RSAKey.import_key(obj)
        new_obj = key.as_dict()
        self.assertBase64IntEqual(new_obj["n"], obj["n"])
        self.assertBase64IntEqual(new_obj["e"], obj["e"])

    def test_rsa_private_key(self):
        # https://tools.ietf.org/html/rfc7520#section-3.4
        obj = read_file_path("jwk_private.json")
        key = RSAKey.import_key(obj)
        new_obj = key.as_dict(is_private=True)
        self.assertBase64IntEqual(new_obj["n"], obj["n"])
        self.assertBase64IntEqual(new_obj["e"], obj["e"])
        self.assertBase64IntEqual(new_obj["d"], obj["d"])
        self.assertBase64IntEqual(new_obj["p"], obj["p"])
        self.assertBase64IntEqual(new_obj["q"], obj["q"])
        self.assertBase64IntEqual(new_obj["dp"], obj["dp"])
        self.assertBase64IntEqual(new_obj["dq"], obj["dq"])
        self.assertBase64IntEqual(new_obj["qi"], obj["qi"])

    def test_rsa_private_key2(self):
        rsa_obj = read_file_path("jwk_private.json")
        obj = {
            "kty": "RSA",
            "kid": "bilbo.baggins@hobbiton.example",
            "use": "sig",
            "n": rsa_obj["n"],
            "d": rsa_obj["d"],
            "e": "AQAB",
        }
        key = RSAKey.import_key(obj)
        new_obj = key.as_dict(is_private=True)
        self.assertBase64IntEqual(new_obj["n"], obj["n"])
        self.assertBase64IntEqual(new_obj["e"], obj["e"])
        self.assertBase64IntEqual(new_obj["d"], obj["d"])
        self.assertBase64IntEqual(new_obj["p"], rsa_obj["p"])
        self.assertBase64IntEqual(new_obj["q"], rsa_obj["q"])
        self.assertBase64IntEqual(new_obj["dp"], rsa_obj["dp"])
        self.assertBase64IntEqual(new_obj["dq"], rsa_obj["dq"])
        self.assertBase64IntEqual(new_obj["qi"], rsa_obj["qi"])

    def test_invalid_rsa(self):
        self.assertRaises(ValueError, RSAKey.import_key, {"kty": "RSA"})
        rsa_obj = read_file_path("jwk_private.json")
        obj = {
            "kty": "RSA",
            "kid": "bilbo.baggins@hobbiton.example",
            "use": "sig",
            "n": rsa_obj["n"],
            "d": rsa_obj["d"],
            "p": rsa_obj["p"],
            "e": "AQAB",
        }
        self.assertRaises(ValueError, RSAKey.import_key, obj)

    def test_rsa_key_generate(self):
        self.assertRaises(ValueError, RSAKey.generate_key, 256)
        self.assertRaises(ValueError, RSAKey.generate_key, 2001)

        key1 = RSAKey.generate_key(is_private=True)
        self.assertIn(b"PRIVATE", key1.as_pem(is_private=True))
        self.assertIn(b"PUBLIC", key1.as_pem(is_private=False))

        key2 = RSAKey.generate_key(is_private=False)
        self.assertRaises(ValueError, key2.as_pem, True)
        self.assertIn(b"PUBLIC", key2.as_pem(is_private=False))


class ECKeyTest(BaseTest):
    def test_ec_public_key(self):
        # https://tools.ietf.org/html/rfc7520#section-3.1
        obj = read_file_path("secp521r1-public.json")
        key = ECKey.import_key(obj)
        new_obj = key.as_dict()
        self.assertEqual(new_obj["crv"], obj["crv"])
        self.assertBase64IntEqual(new_obj["x"], obj["x"])
        self.assertBase64IntEqual(new_obj["y"], obj["y"])
        self.assertEqual(key.as_json()[0], "{")

    def test_ec_private_key(self):
        # https://tools.ietf.org/html/rfc7520#section-3.2
        obj = read_file_path("secp521r1-private.json")
        key = ECKey.import_key(obj)
        new_obj = key.as_dict(is_private=True)
        self.assertEqual(new_obj["crv"], obj["crv"])
        self.assertBase64IntEqual(new_obj["x"], obj["x"])
        self.assertBase64IntEqual(new_obj["y"], obj["y"])
        self.assertBase64IntEqual(new_obj["d"], obj["d"])

    def test_invalid_ec(self):
        self.assertRaises(ValueError, ECKey.import_key, {"kty": "EC"})

    def test_ec_key_generate(self):
        self.assertRaises(ValueError, ECKey.generate_key, "Invalid")

        key1 = ECKey.generate_key("P-384", is_private=True)
        self.assertIn(b"PRIVATE", key1.as_pem(is_private=True))
        self.assertIn(b"PUBLIC", key1.as_pem(is_private=False))

        key2 = ECKey.generate_key("P-256", is_private=False)
        self.assertRaises(ValueError, key2.as_pem, True)
        self.assertIn(b"PUBLIC", key2.as_pem(is_private=False))


class OKPKeyTest(BaseTest):
    def test_import_okp_ssh_key(self):
        raw = read_file_path("ed25519-ssh.pub")
        key = OKPKey.import_key(raw)
        obj = key.as_dict()
        self.assertEqual(obj["kty"], "OKP")
        self.assertEqual(obj["crv"], "Ed25519")

    def test_import_okp_public_key(self):
        obj = {
            "x": "AD9E0JYnpV-OxZbd8aN1t4z71Vtf6JcJC7TYHT0HDbg",
            "crv": "Ed25519",
            "kty": "OKP",
        }
        key = OKPKey.import_key(obj)
        new_obj = key.as_dict()
        self.assertEqual(obj["x"], new_obj["x"])

    def test_import_okp_private_pem(self):
        raw = read_file_path("ed25519-pkcs8.pem")
        key = OKPKey.import_key(raw)
        obj = key.as_dict(is_private=True)
        self.assertEqual(obj["kty"], "OKP")
        self.assertEqual(obj["crv"], "Ed25519")
        self.assertIn("d", obj)

    def test_import_okp_private_dict(self):
        obj = {
            "x": "11qYAYKxCrfVS_7TyWQHOg7hcvPapiMlrwIaaPcHURo",
            "d": "nWGxne_9WmC6hEr0kuwsxERJxWl7MmkZcDusAxyuf2A",
            "crv": "Ed25519",
            "kty": "OKP",
        }
        key = OKPKey.import_key(obj)
        new_obj = key.as_dict(is_private=True)
        self.assertEqual(obj["d"], new_obj["d"])

    def test_okp_key_generate_pem(self):
        self.assertRaises(ValueError, OKPKey.generate_key, "invalid")

        key1 = OKPKey.generate_key("Ed25519", is_private=True)
        self.assertIn(b"PRIVATE", key1.as_pem(is_private=True))
        self.assertIn(b"PUBLIC", key1.as_pem(is_private=False))

        key2 = OKPKey.generate_key("X25519", is_private=False)
        self.assertRaises(ValueError, key2.as_pem, True)
        self.assertIn(b"PUBLIC", key2.as_pem(is_private=False))


class JWKTest(BaseTest):
    def test_generate_keys(self):
        key = JsonWebKey.generate_key(kty="oct", crv_or_size=256, is_private=True)
        self.assertEqual(key["kty"], "oct")

        key = JsonWebKey.generate_key(kty="EC", crv_or_size="P-256")
        self.assertEqual(key["kty"], "EC")

        key = JsonWebKey.generate_key(kty="RSA", crv_or_size=2048)
        self.assertEqual(key["kty"], "RSA")

        key = JsonWebKey.generate_key(kty="OKP", crv_or_size="Ed25519")
        self.assertEqual(key["kty"], "OKP")

    def test_import_keys(self):
        rsa_pub_pem = read_file_path("rsa_public.pem")
        self.assertRaises(ValueError, JsonWebKey.import_key, rsa_pub_pem, {"kty": "EC"})

        key = JsonWebKey.import_key(raw=rsa_pub_pem, options={"kty": "RSA"})
        self.assertIn("e", dict(key))
        self.assertIn("n", dict(key))

        key = JsonWebKey.import_key(raw=rsa_pub_pem)
        self.assertIn("e", dict(key))
        self.assertIn("n", dict(key))

    def test_import_key_set(self):
        jwks_public = read_file_path("jwks_public.json")
        key_set1 = JsonWebKey.import_key_set(jwks_public)
        key1 = key_set1.find_by_kid("abc")
        self.assertEqual(key1["e"], "AQAB")

        key_set2 = JsonWebKey.import_key_set(jwks_public["keys"])
        key2 = key_set2.find_by_kid("abc")
        self.assertEqual(key2["e"], "AQAB")

        key_set3 = JsonWebKey.import_key_set(json_dumps(jwks_public))
        key3 = key_set3.find_by_kid("abc")
        self.assertEqual(key3["e"], "AQAB")

        self.assertRaises(ValueError, JsonWebKey.import_key_set, "invalid")

    def test_thumbprint(self):
        # https://tools.ietf.org/html/rfc7638#section-3.1
        data = read_file_path("thumbprint_example.json")
        key = JsonWebKey.import_key(data)
        expected = "NzbLsXh8uDCcd-6MNwXF4W_7noWXFZAfHkxZsRGC9Xs"
        self.assertEqual(key.thumbprint(), expected)

    def test_key_set(self):
        key = RSAKey.generate_key(is_private=True)
        key_set = KeySet([key])
        obj = key_set.as_dict()["keys"][0]
        self.assertIn("kid", obj)
        self.assertEqual(key_set.as_json()[0], "{")



import time

from authlib.oauth1.rfc5849 import signature
from tests.util import decode_response
from tests.util import read_file_path

from .oauth1_server import Client
from .oauth1_server import TestCase
from .oauth1_server import User
from .oauth1_server import create_authorization_server
from .oauth1_server import db


class TokenCredentialsTest(TestCase):
    USE_CACHE = True

    def prepare_data(self):
        self.server = create_authorization_server(self.app, self.USE_CACHE)
        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="client",
            client_secret="secret",
            default_redirect_uri="https://a.b",
        )
        db.session.add(client)
        db.session.commit()

    def prepare_temporary_credential(self):
        credential = {
            "oauth_token": "abc",
            "oauth_token_secret": "abc-secret",
            "oauth_verifier": "abc-verifier",
            "user": 1,
        }
        func = self.server._hooks["create_temporary_credential"]
        func(credential, "client", "oob")

    def test_invalid_token_request_parameters(self):
        self.prepare_data()
        url = "/oauth/token"

        # case 1
        rv = self.client.post(url)
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_consumer_key", data["error_description"])

        # case 2
        rv = self.client.post(url, data={"oauth_consumer_key": "a"})
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_client")

        # case 3
        rv = self.client.post(url, data={"oauth_consumer_key": "client"})
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_token", data["error_description"])

        # case 4
        rv = self.client.post(
            url, data={"oauth_consumer_key": "client", "oauth_token": "a"}
        )
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_token")

    def test_invalid_token_and_verifiers(self):
        self.prepare_data()
        url = "/oauth/token"
        hook = self.server._hooks["create_temporary_credential"]

        # case 5
        hook(
            {"oauth_token": "abc", "oauth_token_secret": "abc-secret"}, "client", "oob"
        )
        rv = self.client.post(
            url, data={"oauth_consumer_key": "client", "oauth_token": "abc"}
        )
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_verifier", data["error_description"])

        # case 6
        hook(
            {"oauth_token": "abc", "oauth_token_secret": "abc-secret"}, "client", "oob"
        )
        rv = self.client.post(
            url,
            data={
                "oauth_consumer_key": "client",
                "oauth_token": "abc",
                "oauth_verifier": "abc",
            },
        )
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_request")
        self.assertIn("oauth_verifier", data["error_description"])

    def test_duplicated_oauth_parameters(self):
        self.prepare_data()
        url = "/oauth/token?oauth_consumer_key=client"
        rv = self.client.post(
            url,
            data={
                "oauth_consumer_key": "client",
                "oauth_token": "abc",
                "oauth_verifier": "abc",
            },
        )
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "duplicated_oauth_protocol_parameter")

    def test_plaintext_signature(self):
        self.prepare_data()
        url = "/oauth/token"

        # case 1: success
        self.prepare_temporary_credential()
        auth_header = (
            'OAuth oauth_consumer_key="client",'
            'oauth_signature_method="PLAINTEXT",'
            'oauth_token="abc",'
            'oauth_verifier="abc-verifier",'
            'oauth_signature="secret&abc-secret"'
        )
        headers = {"Authorization": auth_header}
        rv = self.client.post(url, headers=headers)
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        # case 2: invalid signature
        self.prepare_temporary_credential()
        rv = self.client.post(
            url,
            data={
                "oauth_consumer_key": "client",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_token": "abc",
                "oauth_verifier": "abc-verifier",
                "oauth_signature": "invalid-signature",
            },
        )
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_signature")

    def test_hmac_sha1_signature(self):
        self.prepare_data()
        url = "/oauth/token"

        params = [
            ("oauth_consumer_key", "client"),
            ("oauth_token", "abc"),
            ("oauth_verifier", "abc-verifier"),
            ("oauth_signature_method", "HMAC-SHA1"),
            ("oauth_timestamp", str(int(time.time()))),
            ("oauth_nonce", "hmac-sha1-nonce"),
        ]
        base_string = signature.construct_base_string(
            "POST", "http://localhost/oauth/token", params
        )
        sig = signature.hmac_sha1_signature(base_string, "secret", "abc-secret")
        params.append(("oauth_signature", sig))
        auth_param = ",".join([f'{k}="{v}"' for k, v in params])
        auth_header = "OAuth " + auth_param
        headers = {"Authorization": auth_header}

        # case 1: success
        self.prepare_temporary_credential()
        rv = self.client.post(url, headers=headers)
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        # case 2: exists nonce
        self.prepare_temporary_credential()
        rv = self.client.post(url, headers=headers)
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_nonce")

    def test_rsa_sha1_signature(self):
        self.prepare_data()
        url = "/oauth/token"

        self.prepare_temporary_credential()
        params = [
            ("oauth_consumer_key", "client"),
            ("oauth_token", "abc"),
            ("oauth_verifier", "abc-verifier"),
            ("oauth_signature_method", "RSA-SHA1"),
            ("oauth_timestamp", str(int(time.time()))),
            ("oauth_nonce", "rsa-sha1-nonce"),
        ]
        base_string = signature.construct_base_string(
            "POST", "http://localhost/oauth/token", params
        )
        sig = signature.rsa_sha1_signature(
            base_string, read_file_path("rsa_private.pem")
        )
        params.append(("oauth_signature", sig))
        auth_param = ",".join([f'{k}="{v}"' for k, v in params])
        auth_header = "OAuth " + auth_param
        headers = {"Authorization": auth_header}
        rv = self.client.post(url, headers=headers)
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        # case: invalid signature
        self.prepare_temporary_credential()
        auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
        auth_header = "OAuth " + auth_param
        headers = {"Authorization": auth_header}
        rv = self.client.post(url, headers=headers)
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_signature")
